import React, { useCallback, useMemo, useRef, useState } from 'react';
import {
  setQuerybuilderNodes,
  useConfig,
  useQuerybuilderGraph,
  useQuerybuilderGraphology,
  useQuerybuilderSettings,
  useSchemaGraphology,
} from '@graphpolaris/shared/lib/data-access/store';
import ReactFlow, {
  Background,
  Connection,
  ControlButton,
  Controls,
  Edge,
  HandleType,
  Node,
  NodeChange,
  NodePositionChange,
  OnConnectStartParams,
  OnEdgesChange,
  ReactFlowInstance,
  ReactFlowProvider,
  XYPosition,
  isNode,
  useReactFlow,
} from 'reactflow';
import styles from './querybuilder.module.scss';

import { clearQB } from '@graphpolaris/shared/lib/data-access/store/querybuilderSlice';
import { Cached as CachedIcon, Delete as DeleteIcon, ImportExport as ExportIcon, Settings as SettingsIcon } from '@mui/icons-material';
import { useDispatch } from 'react-redux';
import {
  AllLogicDescriptions,
  AllLogicMap,
  NodeAttribute,
  QueryElementTypes,
  QueryGraphNodes,
  createReactFlowElements,
  toHandleData,
} from '../model';
import { InputNodeType } from '../model/logic/general';
import { ConnectionDragLine, ConnectionLine, EntityFlowElement, RelationPill } from '../pills';
import LogicPill from '../pills/customFlowPills/logicpill/logicpill';
import { dragPillStarted, movePillTo } from '../pills/dragging/dragPill';
import DifferenceIcon from '@mui/icons-material/Difference';
import LightbulbIcon from '@mui/icons-material/Lightbulb';
import { Dialog } from '../../components/Dialog';
import { QueryBuilderLogicPillsPanel } from './querysidepanel/queryBuilderLogicPillsPanel';
import { QueryBuilderMLPanel } from './querysidepanel/queryBuilderMLPanel';
import { Popup } from '../../components/Popup';
import { QuerySettingsDialog } from './querySettingsDialog';

export type QueryBuilderProps = {
  onRunQuery?: () => void;
};

type SettingsPanel = 'settings' | 'ml' | 'logic' | undefined;

/**
 * This is the main querybuilder component. It is responsible for holding all pills and fire off the visual part of the querybuilder panel logic
 */
export const QueryBuilderInner = (props: QueryBuilderProps) => {
  const [toggleSettings, setToggleSettings] = useState<SettingsPanel>();
  const reactFlowWrapper = useRef<HTMLDivElement>(null);
  const queryBuilderSettings = useQuerybuilderSettings();

  var nodeTypes = {
    entity: EntityFlowElement,
    relation: RelationPill,
    logic: LogicPill,
  };

  var edgeTypes = {
    connection: ConnectionLine,
    attribute_connection: ConnectionLine,
  };

  const schema = useSchemaGraphology();
  const graphologyGraph = useQuerybuilderGraphology();
  const graph = useQuerybuilderGraph();
  const config = useConfig();
  const dispatch = useDispatch();
  const isDraggingPill = useRef(false);
  const connectingNodeId = useRef<{
    params: OnConnectStartParams;
    position: XYPosition;
    node: QueryGraphNodes;
    attribute: NodeAttribute;
  } | null>(null);
  const reactFlow = useReactFlow();
  const isEdgeUpdating = useRef(false);
  const isOnConnect = useRef(false);

  const onInit = (reactFlowInstance: ReactFlowInstance) => {
    setTimeout(() => reactFlow.fitView(), 0);
  };

  const elements = useMemo(() => createReactFlowElements(graphologyGraph), [graph]);

  /**
   * Called when a node is dragged in querybuilder to allow for movement
   * @param event
   * @param node
   */
  const onNodeDrag = (event: React.MouseEvent<Element, MouseEvent>, node: Node<any>, nodes: Node[]) => {
    // console.log(nodes);
    // // Get the node in the elements list to get the previous location
    // const pNode = elements.nodes.find((e) => e?.id === node?.id);
    // if (!(pNode && isNode(pNode))) return;
    // // This is then used to calculate the delta position
    // const dx = node.position.x - pNode.position.x;
    // const dy = node.position.y - pNode.position.y;
    // // Check if we started dragging, if so, call the drag started usecase
    // if (!isDraggingPill.current) {
    //   dragPillStarted(node.id, graphologyGraph);
    //   isDraggingPill.current = true;
    // }
    // // Call the drag usecase
    // movePillTo(node.id, graphologyGraph, dx, dy, node.position);
    // // Dispatch the new graphology object, so reactflow will get rerendered
    // dispatch(setQuerybuilderNodes(graphologyGraph.export()));
  };

  /**
   * Called when a node is released from dragging in querybuilder
   * @param event
   * @param node
   **/
  const onNodeDragStop = (event: React.MouseEvent<Element, MouseEvent>, node: Node<any>) => {
    // isDraggingPill.current = false;
    //
    // // Call the drag pill stopped usecase
    // dragPillStopped(node.id, graphologyGraph);
    //
    // // Dispatch the new graphology object, so reactflow will get rerendered
    // dispatch(setQuerybuilderNodes(graphologyGraph.export()));
  };

  /**
   * Clears all nodes in the graph.
   */
  function clearAllNodes() {
    dispatch(clearQB());
  }

  /**
   * Clears all nodes in the graph.
   *  TODO: only works if the node is clicked and not moved (maybe use onSelectionChange)
   */
  function onNodesDelete(nodes: Node[]) {
    nodes.forEach((n) => {
      graphologyGraph.dropNode(n.id);
    });

    // Dispatch the new graphology object, so reactflow will get re-rendered
    dispatch(setQuerybuilderNodes(graphologyGraph.export()));
  }

  /**
   * TODO?
   */
  function onNodesChange(nodes: NodeChange[]) {
    // console.log(nodes);
    nodes.forEach((n) => {
      // console.log(graphologyGraph.findNode((gn) => gn === n?.id));
      if (n.type !== 'position') {
        // updated = true;
        // graphologyGraph.updateAttributes(n.id, n.data);
      } else {
        const node = n as NodePositionChange;
        // Get the node in the elements list to get the previous location
        const pNode = elements.nodes.find((e) => e?.id === node?.id);
        if (!(pNode && isNode(pNode)) || !node?.position) return;
        // This is then used to calculate the delta position
        const dx = node.position.x - pNode.position.x;
        const dy = node.position.y - pNode.position.y;
        // Check if we started dragging, if so, call the drag started usecase
        if (!isDraggingPill.current) {
          dragPillStarted(node.id, graphologyGraph);
          isDraggingPill.current = true;
        }
        // Call the drag usecase
        movePillTo(node.id, graphologyGraph, dx, dy, node.position);
        // Dispatch the new graphology object, so reactflow will get rerendered
        dispatch(setQuerybuilderNodes(graphologyGraph.export()));
      }
    });
  }

  const onDragOver = (event: React.DragEvent<HTMLDivElement>): void => {
    event.preventDefault();
    event.dataTransfer.dropEffect = 'move';
  };

  /**
   * The onDrop is called when the user drops an element from the schema onto the QueryBuilder.
   * In the onDrop query elements will be created based on the data stored in the drag event (datastrasfer).
   * @param event Drag event.
   */
  const onDrop = (event: React.DragEvent<HTMLDivElement>): void => {
    event.preventDefault();

    // The dropped element should be a valid reactflow element
    const data: string = event.dataTransfer.getData('application/reactflow');
    if (data.length == 0 || !reactFlow) return;

    const dragData = JSON.parse(data);
    const bounds = reactFlowWrapper?.current?.getBoundingClientRect() || { x: 0, y: 0 };
    const position = reactFlow.project({
      //TODO: this position should be centre of entity, rather than topleft
      x: event.clientX - bounds.x,
      y: event.clientY - bounds.y,
    });

    switch (dragData.type) {
      case QueryElementTypes.Entity:
        // console.log('entity drop', dragData, schema.getNodeAttribute(dragData.name, 'attributes'));
        graphologyGraph.addPill2Graphology(
          {
            type: QueryElementTypes.Entity,
            x: position.x,
            y: position.y,
            name: dragData.name,
          },
          schema.getNodeAttribute(dragData.name, 'attributes')
        );
        dispatch(setQuerybuilderNodes(graphologyGraph.export()));
        break;
      // Creates a relation element and will also create the 2 related entities together with the connections
      case QueryElementTypes.Relation:
        const relation = graphologyGraph.addPill2Graphology(
          {
            type: QueryElementTypes.Relation,
            x: position.x,
            y: position.y,
            depth: { min: queryBuilderSettings.depth.min, max: queryBuilderSettings.depth.max },
            name: dragData.collection,
            collection: dragData.collection,
          },
          schema.getEdgeAttribute(dragData.label, 'attributes')
        );

        if (config.autoSendQueries) {
          // sendQuery();
        }

        dispatch(setQuerybuilderNodes(graphologyGraph.export()));
        break;
      default:
        const logic = AllLogicMap[dragData.value.key];
        const firstLeftLogicInput = logic.inputs?.[0];
        if (!firstLeftLogicInput) return;

        const logicNode = graphologyGraph.addLogicPill2Graphology({
          name: dragData.value.name,
          type: QueryElementTypes.Logic,
          x: position.x,
          y: position.y,
          logic: logic,
        });

        dispatch(setQuerybuilderNodes(graphologyGraph.export()));
    }
  };

  const onNodeMouseEnter = (event: React.MouseEvent, node: Node) => {
    console.log(node);
    console.log(schema.getNodeAttribute(node.type, 'attributes'));
  };

  const onNodeMouseLeave = (event: React.MouseEvent, node: Node) => {
    console.log(node);
  };

  const onConnect = useCallback(
    (connection: Connection) => {
      if (!isEdgeUpdating.current) {
        isOnConnect.current = true;
        if (!connection.sourceHandle || !connection.targetHandle) throw new Error('Connection has no source or target');
        console.log('onConnect', connection);

        if (!graphologyGraph.hasEdge(connection.source, connection.target)) {
          graphologyGraph.addEdge(connection.source, connection.target, {
            type: 'connection',
            sourceHandleData: toHandleData(connection.sourceHandle),
            targetHandleData: toHandleData(connection.targetHandle),
          });
          dispatch(setQuerybuilderNodes(graphologyGraph.export()));
        }
      }
    },
    [graph]
  );

  const onConnectStart = useCallback(
    (event: React.MouseEvent | React.TouchEvent, params: OnConnectStartParams) => {
      // console.log('onConnectStart', params);
      if (!params?.handleId) return;

      let node = graphologyGraph.getNodeAttributes(params.nodeId);
      const handleData = toHandleData(params.handleId);

      connectingNodeId.current = {
        params,
        node,
        position: { x: 0, y: 0 },
        attribute: { handleData: handleData },
      };
    },
    [graph]
  );

  const onConnectEnd = useCallback(
    (event: any) => {
      const targetIsPane = event.target.classList.contains('react-flow__pane');
      if (isOnConnect.current) {
        isOnConnect.current = false;
        return;
      }

      if (targetIsPane && !isEdgeUpdating.current) {
        let clientX: number = 0;
        let clientY: number = 0;

        if ('touches' in event) clientX = event?.touches?.[0]?.clientX;
        else if ('clientX' in event) clientX = event?.clientX;
        if ('touches' in event) clientY = event?.touches?.[0]?.clientY;
        else if ('clientY' in event) clientY = event?.clientY;

        if (reactFlowWrapper.current) {
          const { top, left } = reactFlowWrapper.current.getBoundingClientRect();
          clientX -= left;
          clientY -= top;
        }
        const position = reactFlow.project({ x: clientX, y: clientY });
        if (connectingNodeId?.current) connectingNodeId.current.position = position;

        setToggleSettings('logic');
      }
    },
    [reactFlow.project]
  );

  const onNewNodeFromPopup = (value: AllLogicDescriptions) => {
    console.log('onNewNodeFromPopup', value);

    const logic = AllLogicMap[value.key];
    const firstLeftLogicInput = logic.inputs?.[0];
    if (!firstLeftLogicInput) return;

    if (connectingNodeId.current === null || connectingNodeId.current?.params?.handleId == null) {
      const bounds = reactFlowWrapper?.current?.getBoundingClientRect() || { x: 0, y: 0, width: 0, height: 0 };

      const logicNode = graphologyGraph.addLogicPill2Graphology({
        name: value.name,
        type: QueryElementTypes.Logic,
        x: bounds.width / 2,
        y: bounds.height / 2,
        logic: logic,
      });
    } else {
      const params = connectingNodeId.current.params;
      const position = connectingNodeId.current.position;

      // console.log('onNewNodeFromPopup', value, type, params, position);

      const logicNode = graphologyGraph.addLogicPill2Graphology({
        name: value.name,
        type: QueryElementTypes.Logic,
        x: position.x,
        y: position.y,
        logic: logic,
      });

      if (!logicNode?.id) throw new Error('Logic node has no id');
      if (!logicNode?.name) throw new Error('Logic node has no name');
      if (!params.handleId) throw new Error('Connection has no source or target');

      const sourceHandleData = toHandleData(params.handleId);
      graphologyGraph.addEdge2Graphology(
        graphologyGraph.getNodeAttributes(params.nodeId),
        graphologyGraph.getNodeAttributes(logicNode.id),
        { type: 'connection' },
        { sourceHandleName: sourceHandleData.attributeName, targetHandleName: firstLeftLogicInput.name }
      );
    }

    dispatch(setQuerybuilderNodes(graphologyGraph.export()));
    setToggleSettings(undefined);
    connectingNodeId.current = null;
  };

  const onEdgeUpdateStart = useCallback(() => {
    isEdgeUpdating.current = true;
  }, []);

  const onEdgeUpdate = useCallback(
    (oldEdge: Edge, newConnection: Connection) => {
      if (isEdgeUpdating.current) {
        isEdgeUpdating.current = false;

        if (graphologyGraph.hasEdge(oldEdge.id)) {
          graphologyGraph.dropEdge(oldEdge.id);
        }

        if (!newConnection.sourceHandle || !newConnection.targetHandle) throw new Error('Connection has no source or target');
        graphologyGraph.addEdge(newConnection.source, newConnection.target, {
          type: 'connection',
          sourceHandleData: toHandleData(newConnection.sourceHandle),
          targetHandleData: toHandleData(newConnection.targetHandle),
        });
        dispatch(setQuerybuilderNodes(graphologyGraph.export()));
      }
    },
    [graph]
  );

  const onEdgesChange = (params: OnEdgesChange) => {
    console.log('edges change', params);
  };

  const onEdgeUpdateEnd = useCallback(
    (event: MouseEvent | TouchEvent, edge: Edge, handleType: HandleType) => {
      if (isEdgeUpdating.current) {
        if (graphologyGraph.hasEdge(edge.id)) {
          graphologyGraph.dropEdge(edge.id);
        }
        dispatch(setQuerybuilderNodes(graphologyGraph.export()));
      }
      isEdgeUpdating.current = false;
    },
    [graph]
  );

  const onNodeContextMenu = (event: React.MouseEvent, node: Node) => {
    event.preventDefault();
    graphologyGraph.dropNode(node.id);
    dispatch(setQuerybuilderNodes(graphologyGraph.export()));
  };

  return (
    <div ref={reactFlowWrapper} className="h-full w-full">
      <QuerySettingsDialog open={toggleSettings === 'settings'} onClose={() => setToggleSettings(undefined)} />

      <Dialog
        open={toggleSettings === 'logic'}
        onClose={() => {
          setToggleSettings(undefined);
        }}
      >
        <QueryBuilderLogicPillsPanel
          onClick={onNewNodeFromPopup}
          title="Logic Pills usable by the node"
          className="min-h-[75vh] max-h-[75vh]"
          onDrag={() => {}}
        />
      </Dialog>
      <Popup open={toggleSettings === 'ml'} hAnchor="right">
        <QueryBuilderMLPanel />
      </Popup>
      <ReactFlow
        edges={elements.edges}
        nodes={elements.nodes}
        snapGrid={[10, 10]}
        snapToGrid
        nodeTypes={nodeTypes}
        edgeTypes={edgeTypes}
        connectionLineComponent={ConnectionDragLine}
        // connectionMode={ConnectionMode.Loose}
        onInit={onInit}
        onNodesChange={onNodesChange}
        onNodeDrag={onNodeDrag}
        onNodeDragStop={onNodeDragStop}
        onDragOver={onDragOver}
        onConnect={onConnect}
        onConnectStart={onConnectStart}
        onConnectEnd={onConnectEnd}
        // onNodeMouseEnter={onNodeMouseEnter}
        // onNodeMouseLeave={onNodeMouseLeave}
        onEdgeUpdate={onEdgeUpdate}
        onEdgeUpdateStart={onEdgeUpdateStart}
        onEdgeUpdateEnd={onEdgeUpdateEnd}
        onDrop={onDrop}
        // onContextMenu={onContextMenu}
        onNodeContextMenu={onNodeContextMenu}
        // onNodesDelete={onNodesDelete}
        // onNodesChange={onNodesChange}
        deleteKeyCode="Backspace"
        className={styles.reactflow}
        attributionPosition="top-right"
      >
        <Background gap={10} size={0.7} />

        <Controls showZoom={false} showInteractive={false} className={styles.controls}>
          <ControlButton className={styles.buttons} title={'Remove all elements'} onClick={() => clearAllNodes()}>
            <DeleteIcon />
          </ControlButton>
          <ControlButton
            className={styles.buttons}
            title={'Export querybuilder'}
            onClick={(event) => {
              event.stopPropagation();
            }}
          >
            <ExportIcon />
          </ControlButton>
          <ControlButton
            className={styles.buttons + (toggleSettings === 'settings' ? ' btn-active' : '')}
            title={'Other settings'}
            onClick={(event) => {
              event.stopPropagation();
              if (toggleSettings === 'settings') setToggleSettings(undefined);
              else setToggleSettings('settings');
            }}
          >
            <SettingsIcon />
          </ControlButton>
          <ControlButton
            className={styles.buttons}
            title={'Re-Run Query'}
            onClick={(event) => {
              event.stopPropagation();
              if (props.onRunQuery) props.onRunQuery();
            }}
          >
            <CachedIcon />
          </ControlButton>
          <ControlButton
            className={styles.buttons + (toggleSettings === 'logic' ? ' btn-active' : '')}
            title={'Logic Pills'}
            onClick={(event) => {
              event.stopPropagation();
              if (toggleSettings === 'logic') setToggleSettings(undefined);
              else setToggleSettings('logic');
            }}
          >
            <DifferenceIcon />
          </ControlButton>
          <ControlButton
            className={styles.buttons + (toggleSettings === 'ml' ? ' btn-active' : '')}
            title={'Machine Learning'}
            onClick={(event) => {
              event.stopPropagation();
              if (toggleSettings === 'ml') setToggleSettings(undefined);
              else setToggleSettings('ml');
            }}
          >
            <LightbulbIcon />
          </ControlButton>
        </Controls>
      </ReactFlow>
    </div>
  );
};

export const QueryBuilder = (props: QueryBuilderProps) => {
  return (
    <div className="flex w-full h-full">
      <ReactFlowProvider>
        <QueryBuilderInner {...props} />
      </ReactFlowProvider>
      {/* <QuerySidePanel title="Query Panel" draggable /> */}
    </div>
  );
};

export default QueryBuilder;
